# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from decimal import Decimal
import json
import binascii

from magnetodb.common import exception
from magnetodb.common.exception import ConditionalCheckFailedException

from magnetodb.openstack.common import log as logging
from magnetodb.storage import models
from magnetodb.storage.driver import StorageDriver

LOG = logging.getLogger(__name__)

STORAGE_TO_CASSANDRA_TYPES = {
    models.ATTRIBUTE_TYPE_STRING: 'text',
    models.ATTRIBUTE_TYPE_NUMBER: 'decimal',
    models.ATTRIBUTE_TYPE_BLOB: 'blob',
    models.ATTRIBUTE_TYPE_STRING_SET: 'set<text>',
    models.ATTRIBUTE_TYPE_NUMBER_SET: 'set<decimal>',
    models.ATTRIBUTE_TYPE_BLOB_SET: 'set<blob>'
}

CASSANDRA_TO_STORAGE_TYPES = {val: key for key, val
                              in STORAGE_TO_CASSANDRA_TYPES.iteritems()}

CONDITION_TO_OP = {
    models.Condition.CONDITION_TYPE_EQUAL: '=',
    models.IndexedCondition.CONDITION_TYPE_LESS: '<',
    models.IndexedCondition.CONDITION_TYPE_LESS_OR_EQUAL: '<=',
    models.IndexedCondition.CONDITION_TYPE_GREATER: '>',
    models.IndexedCondition.CONDITION_TYPE_GREATER_OR_EQUAL: '>=',
}

USER_PREFIX = 'user_'
USER_PREFIX_LENGTH = len(USER_PREFIX)

SYSTEM_KEYSPACE = 'magnetodb'
SYSTEM_COLUMN_INDEX_NAME = 'index_name'
SYSTEM_COLUMN_INDEX_VALUE_STRING = 'index_value_string'
SYSTEM_COLUMN_INDEX_VALUE_NUMBER = 'index_value_number'
SYSTEM_COLUMN_INDEX_VALUE_BLOB = 'index_value_blob'

LOCAL_INDEX_FIELD_LIST = [
    SYSTEM_COLUMN_INDEX_NAME,
    SYSTEM_COLUMN_INDEX_VALUE_STRING,
    SYSTEM_COLUMN_INDEX_VALUE_NUMBER,
    SYSTEM_COLUMN_INDEX_VALUE_BLOB
]

INDEX_TYPE_TO_INDEX_POS_MAP = {
    models.ATTRIBUTE_TYPE_STRING: 1,
    models.ATTRIBUTE_TYPE_NUMBER: 2,
    models.ATTRIBUTE_TYPE_BLOB: 3,
}

SYSTEM_COLUMN_EXTRA_ATTR_DATA = 'extra_attr_data'
SYSTEM_COLUMN_EXTRA_ATTR_TYPES = 'extra_attr_types'
SYSTEM_COLUMN_ATTR_EXIST = 'attr_exist'

DEFAULT_STRING_VALUE = models.AttributeValue.str('')
DEFAULT_NUMBER_VALUE = models.AttributeValue.number(0)
DEFAULT_BLOB_VALUE = models.AttributeValue.blob('')


def _encode_predefined_attr_value(attr_value):
    if attr_value is None:
        return 'null'
    if attr_value.type.collection_type:
        values = ','.join(map(
            lambda el: _encode_single_value_as_predefined_attr(
                el, attr_value.type.element_type),
            attr_value.value
        ))
        return '{{{}}}'.format(values)
    else:
        return _encode_single_value_as_predefined_attr(
            attr_value.value, attr_value.type.element_type
        )


def _encode_single_value_as_predefined_attr(value, element_type):
    if element_type == models.AttributeType.ELEMENT_TYPE_STRING:
        return "'{}'".format(value)
    elif element_type == models.AttributeType.ELEMENT_TYPE_NUMBER:
        return str(value)
    elif element_type == models.AttributeType.ELEMENT_TYPE_BLOB:
        return "0x{}".format(binascii.hexlify(value))
    else:
        assert False, "Value wasn't formatted for cql query"


def _encode_dynamic_attr_value(attr_value):
    if attr_value is None:
        return 'null'

    val = attr_value.value
    if attr_value.type.collection_type:
        val = map(
            lambda el: _encode_single_value_as_dynamic_attr(
                el, attr_value.type.element_type
            ),
            val
        )
        val.sort()
    else:
        val = _encode_single_value_as_dynamic_attr(
            val, attr_value.type.element_type)
    return "0x{}".format(binascii.hexlify(json.dumps(val)))


def _encode_single_value_as_dynamic_attr(value, element_type):
    if element_type == models.AttributeType.ELEMENT_TYPE_STRING:
        return value
    elif element_type == models.AttributeType.ELEMENT_TYPE_NUMBER:
        return str(value)
    elif element_type == models.AttributeType.ELEMENT_TYPE_BLOB:
        return value
    else:
        assert False, "Value wasn't formatted for cql query"


def _decode_predefined_attr(table_info, cas_name, cas_val, prefix=USER_PREFIX):
    assert cas_name.startswith(prefix) and cas_val

    name = cas_name[len(USER_PREFIX):]
    storage_type = table_info.schema.attribute_type_map[name]
    return name, models.AttributeValue(storage_type, cas_val)


def _decode_dynamic_value(value, storage_type):
    value = json.loads(value)

    return models.AttributeValue(storage_type, value)


def _decode_single_value(value, element_type):
    if element_type == models.AttributeType.ELEMENT_TYPE_STRING:
        return value
    elif element_type == models.AttributeType.ELEMENT_TYPE_NUMBER:
        return Decimal(value)
    elif element_type == models.AttributeType.ELEMENT_TYPE_BLOB:
        return value
    else:
        assert False, "Value wasn't formatted for cql query"

ENCODED_DEFAULT_STRING_VALUE = _encode_predefined_attr_value(
    DEFAULT_STRING_VALUE
)
ENCODED_DEFAULT_NUMBER_VALUE = _encode_predefined_attr_value(
    DEFAULT_NUMBER_VALUE
)
ENCODED_DEFAULT_BLOB_VALUE = _encode_predefined_attr_value(
    DEFAULT_BLOB_VALUE
)


class CassandraStorageDriver(StorageDriver):
    def __init__(self, cluster_handler, table_info_repo):
        self.__cluster_handler = cluster_handler
        self.__table_info_repo = table_info_repo

    def create_table(self, context, table_name):
        table_info = self.__table_info_repo.get(context, table_name)
        table_schema = table_info.schema

        cas_table_name = USER_PREFIX + table_name
        cas_keyspace = USER_PREFIX + context.tenant

        key_count = len(table_schema.key_attributes)

        if key_count < 1 or key_count > 2:
            raise exception.BackendInteractionException(
                "Expected 1 or 2 key attribute(s). Found {}: {}".format(
                    key_count, table_schema.key_attributes))

        hash_key_name = table_schema.key_attributes[0]
        range_key_name = (
            table_schema.key_attributes[1] if key_count > 1 else None
        )

        query_builder = [
            'CREATE TABLE "', cas_keyspace, '"."', cas_table_name, '" ('
        ]

        if table_schema.index_def_map:
            query_builder += (
                SYSTEM_COLUMN_INDEX_NAME, " text,",
                SYSTEM_COLUMN_INDEX_VALUE_STRING, " text, ",
                SYSTEM_COLUMN_INDEX_VALUE_NUMBER, " decimal, ",
                SYSTEM_COLUMN_INDEX_VALUE_BLOB, " blob, ",
            )

        for attr_name, attr_type in (
                table_schema.attribute_type_map.iteritems()):

            query_builder += (
                '"', USER_PREFIX, attr_name, '" ',
                STORAGE_TO_CASSANDRA_TYPES[attr_type], ","
            )

        query_builder += (
            SYSTEM_COLUMN_EXTRA_ATTR_DATA, " map<text, blob>,",
            SYSTEM_COLUMN_EXTRA_ATTR_TYPES, " map<text, text>,",
            SYSTEM_COLUMN_ATTR_EXIST, " set<text>,"
            'PRIMARY KEY ("',
            USER_PREFIX, hash_key_name, '"'
        )

        if table_schema.index_def_map:
            query_builder += (
                ",", SYSTEM_COLUMN_INDEX_NAME,
                ",", SYSTEM_COLUMN_INDEX_VALUE_STRING,
                ",", SYSTEM_COLUMN_INDEX_VALUE_NUMBER,
                ",", SYSTEM_COLUMN_INDEX_VALUE_BLOB,
            )

        if range_key_name:
            query_builder += (
                ',"', USER_PREFIX, range_key_name, '"'
            )

        query_builder.append("))")

        self.__cluster_handler.execute_query("".join(query_builder))
        LOG.debug("Create Table CQL request executed. "
                  "Waiting for schema agreement...")

        self.__cluster_handler.wait_for_table_status(
            keyspace_name=cas_keyspace, table_name=cas_table_name,
            expected_exists=True)
        LOG.debug("Waiting for schema agreement... Done")

        table_info.internal_name = cas_table_name
        self.__table_info_repo.update(
            context, table_info, ["internal_name"]
        )

    def delete_table(self, context, table_name):
        table_info = self.__table_info_repo.get(context, table_name)

        cas_table_name = table_info.internal_name
        cas_keyspace_name = USER_PREFIX + context.tenant

        query = 'DROP TABLE "{}"."{}"'.format(cas_keyspace_name,
                                              cas_table_name)

        self.__cluster_handler.execute_query(query)

        LOG.debug("Delete Table CQL request executed. "
                  "Waiting for schema agreement...")

        self.__cluster_handler.wait_for_table_status(
            keyspace_name=cas_keyspace_name, table_name=cas_table_name,
            expected_exists=False
        )

        LOG.debug("Waiting for schema agreement... Done")

    @staticmethod
    def _append_types_system_attr_value(table_schema, attribute_map,
                                        query_builder=None, prefix=""):
        if query_builder is None:
            query_builder = []
        query_builder.append(prefix)
        prefix = ""
        query_builder.append("{")
        for attr, val in attribute_map.iteritems():
            if (val is not None) and (
                    attr not in table_schema.attribute_type_map):
                query_builder += (
                    prefix, "'", attr, "':'",
                    STORAGE_TO_CASSANDRA_TYPES[val.type], "'"
                )
                prefix = ","
        query_builder.append("}")
        return query_builder

    @staticmethod
    def _append_exists_system_attr_value(attribute_map, query_builder=None,
                                         prefix=""):
        if query_builder is None:
            query_builder = []
        query_builder.append(prefix)
        prefix = ""
        query_builder.append("{")
        for attr, _ in attribute_map.iteritems():
            query_builder += (prefix, "'", attr, "'")
            prefix = ","
        query_builder.append("}")
        return query_builder

    def _append_insert_query(
            self, table_info, attribute_map, query_builder=None,
            index_name=None, index_value=None, if_not_exists=False):
        if query_builder is None:
            query_builder = []

        query_builder += (
            'INSERT INTO "', table_info.internal_keyspace, '"."',
            table_info.internal_name, '" ('
        )
        attr_values = []
        dynamic_attr_names = []
        dynamic_attr_values = []
        for name, val in attribute_map.iteritems():
            if name in table_info.schema.attribute_type_map.keys():
                query_builder += (
                    '"', USER_PREFIX, name, '",'
                )
                attr_values.append(_encode_predefined_attr_value(val))
            else:
                dynamic_attr_names.append(name)
                dynamic_attr_values.append(
                    _encode_dynamic_attr_value(val)
                )

        if table_info.schema.index_def_map:
            query_builder += (
                SYSTEM_COLUMN_INDEX_NAME, ",",
                SYSTEM_COLUMN_INDEX_VALUE_STRING, ",",
                SYSTEM_COLUMN_INDEX_VALUE_NUMBER, ",",
                SYSTEM_COLUMN_INDEX_VALUE_BLOB, ",",
            )

        query_builder += (
            SYSTEM_COLUMN_EXTRA_ATTR_DATA, ",",
            SYSTEM_COLUMN_EXTRA_ATTR_TYPES, ",",
            SYSTEM_COLUMN_ATTR_EXIST,
            ") VALUES("
        )

        for attr_value in attr_values:
            query_builder += (
                attr_value, ","
            )

        if table_info.schema.index_def_map:
            res_index_values = [
                ENCODED_DEFAULT_STRING_VALUE,
                ENCODED_DEFAULT_STRING_VALUE,
                ENCODED_DEFAULT_NUMBER_VALUE,
                ENCODED_DEFAULT_BLOB_VALUE
            ]

            if index_name:
                res_index_values[0] = _encode_single_value_as_predefined_attr(
                    index_name, models.AttributeType.ELEMENT_TYPE_STRING
                )

                res_index_values[INDEX_TYPE_TO_INDEX_POS_MAP[
                    index_value.type]
                ] = _encode_predefined_attr_value(index_value)

            for value in res_index_values:
                query_builder += (
                    value, ","
                )

        query_builder.append("{")

        if dynamic_attr_values:
            dynamic_value_iter = iter(dynamic_attr_values)
            for name in dynamic_attr_names:
                query_builder += (
                    "'", name, "':" + dynamic_value_iter.next(), ","
                )
            query_builder.pop()

        query_builder.append("},")
        self._append_types_system_attr_value(table_info.schema, attribute_map,
                                             query_builder)
        self._append_exists_system_attr_value(attribute_map, query_builder,
                                              prefix=",")
        query_builder.append(")")

        if if_not_exists:
            query_builder.append(" IF NOT EXISTS")
        return query_builder

    def _append_update_query_with_basic_pk(self, table_info, attribute_map,
                                           query_builder=None, rewrite=False):
        if query_builder is None:
            query_builder = []

        key_attr_names = table_info.schema.key_attributes

        not_processed_predefined_attr_names = set(
            table_info.schema.attribute_type_map.keys()
        )

        query_builder += (
            'UPDATE "', table_info.internal_keyspace, '"."',
            table_info.internal_name, '" SET '
        )

        dynamic_attrs_to_set = []
        dynamic_attrs_to_delete = []

        set_prefix = ""

        for name, val in attribute_map.iteritems():
            if name in key_attr_names:
                not_processed_predefined_attr_names.remove(name)
            elif name in not_processed_predefined_attr_names:
                query_builder += (
                    set_prefix, '"', USER_PREFIX, name, '"=',
                    _encode_predefined_attr_value(val),
                )
                set_prefix = ","
                not_processed_predefined_attr_names.remove(name)
            else:
                if val is None:
                    dynamic_attrs_to_delete.append(name)
                else:
                    dynamic_attrs_to_set.append(
                        (name, _encode_dynamic_attr_value(val))
                    )

        if rewrite:
            query_builder += (set_prefix, SYSTEM_COLUMN_EXTRA_ATTR_DATA, "={")

            field_prefix = ""
            for name, value in dynamic_attrs_to_set:
                query_builder += (
                    field_prefix, "'", name, "':", value
                )
                field_prefix = ","
            query_builder.append("},")

            for name in not_processed_predefined_attr_names:
                query_builder += (
                    '"', USER_PREFIX, name, '"=null,'
                )

            query_builder.append(SYSTEM_COLUMN_EXTRA_ATTR_TYPES)
            self._append_types_system_attr_value(table_info.schema,
                                                 attribute_map, query_builder,
                                                 prefix="=")
            query_builder += (",", SYSTEM_COLUMN_ATTR_EXIST)
            self._append_exists_system_attr_value(attribute_map, query_builder,
                                                  prefix="=")
        else:
            if dynamic_attrs_to_set:
                query_builder += (
                    set_prefix, SYSTEM_COLUMN_EXTRA_ATTR_DATA, "=",
                    SYSTEM_COLUMN_EXTRA_ATTR_DATA, "+{"
                )
                set_prefix = ","

                field_prefix = ""
                for name, value in dynamic_attrs_to_set:
                    query_builder += (
                        field_prefix, "'", name, "':", value
                    )
                    field_prefix = ","
                query_builder.append("},")

                query_builder += (
                    SYSTEM_COLUMN_EXTRA_ATTR_TYPES, "=",
                    SYSTEM_COLUMN_EXTRA_ATTR_TYPES
                )
                self._append_types_system_attr_value(
                    table_info.schema, attribute_map, query_builder, prefix="+"
                )
                query_builder += (
                    ",", SYSTEM_COLUMN_ATTR_EXIST, "=",
                    SYSTEM_COLUMN_ATTR_EXIST
                )
                self._append_exists_system_attr_value(
                    attribute_map, query_builder, prefix="+"
                )
            if dynamic_attrs_to_delete:
                for name in dynamic_attrs_to_delete:
                    query_builder += (
                        set_prefix,
                        SYSTEM_COLUMN_EXTRA_ATTR_DATA, "['", name, "']=null,",
                        SYSTEM_COLUMN_EXTRA_ATTR_TYPES, "['", name, "']=null,",
                    )
                query_builder += (
                    SYSTEM_COLUMN_ATTR_EXIST, "=",
                    SYSTEM_COLUMN_ATTR_EXIST, "-{"
                )
                field_prefix = ""
                for name in dynamic_attrs_to_delete:
                    query_builder += (
                        field_prefix, "'", name, "'",
                    )
                    field_prefix = ","
                query_builder.append("}")
        self._append_primary_key(table_info.schema, attribute_map,
                                 query_builder)

        return query_builder

    def _append_update_query(self, table_info, attribute_map,
                             query_builder=None, index_name=None,
                             index_value=None, expected_condition_map=None,
                             rewrite=False):
        query_builder = self._append_update_query_with_basic_pk(
            table_info, attribute_map, query_builder, rewrite=rewrite
        )

        if table_info.schema.index_def_map:
            self._append_index_extra_primary_key(query_builder, index_name,
                                                 index_value, " AND ")

        if expected_condition_map:
            self._append_expected_conditions(
                expected_condition_map, table_info.schema,
                query_builder
            )

        return query_builder

    def _append_update_indexes_queries(self, table_info, old_attribute_map,
                                       attribute_map, query_builder=None,
                                       separator=" ", rewrite=False):
        if query_builder is None:
            query_builder = []
        base_update_query = None
        base_delete_query = None

        def create_base_update_query():
            base_query_builder = (
                self._append_update_query_with_basic_pk(
                    table_info, attribute_map, rewrite=rewrite
                )
            )
            return "".join(base_query_builder)

        def create_base_delete_query():
            base_query_builder = (
                self._append_delete_query_with_basic_pk(
                    table_info, attribute_map
                )
            )
            return "".join(base_query_builder)

        for index_name, index_def in (
                table_info.schema.index_def_map.iteritems()):
            new_index_value = attribute_map.get(
                index_def.attribute_to_index, None
            )
            old_index_value = old_attribute_map.get(
                index_def.attribute_to_index, None
            )
            if new_index_value:
                base_update_query = (
                    base_update_query or create_base_update_query()
                )
                query_builder += (separator, base_update_query)
                self._append_index_extra_primary_key(
                    query_builder, index_name, new_index_value,
                )
            if old_index_value and old_index_value != new_index_value:
                base_delete_query = (
                    base_delete_query or create_base_delete_query()
                )
                query_builder += (separator, base_delete_query)
                self._append_index_extra_primary_key(
                    query_builder, index_name, old_index_value,
                )
        return query_builder

    def _put_item_if_not_exists(self, table_info, attribute_map):
        query_builder = self._append_insert_query(
            table_info, attribute_map, if_not_exists=True
        )

        if table_info.schema.index_def_map:
            qb_len = len(query_builder)
            self._append_update_indexes_queries(
                table_info, {}, attribute_map, query_builder, rewrite=True
            )
            if len(query_builder) > qb_len:
                query_builder.insert(0, self._get_batch_begin_clause())
                query_builder.append(self._get_batch_apply_clause())

            result = self.__cluster_handler.execute_query(
                "".join(query_builder), consistent=True)

            return result[0]['[applied]']

    def put_item(self, context, put_request, if_not_exist=False,
                 expected_condition_map=None):
        """
        @param context: current request context
        @param put_request: contains PutItemRequest items to perform
                    put item operation
        @param if_not_exist: put item only is row is new record (It is possible
                    to use only one of if_not_exist and expected_condition_map
                    parameter)
        @param expected_condition_map: expected attribute name to
                    ExpectedCondition instance mapping. It provides
                    preconditions to make decision about should item be put or
                    not

        @return: True if operation performed, otherwise False

        @raise BackendInteractionException
        """

        table_info = self.__table_info_repo.get(context,
                                                put_request.table_name)

        if if_not_exist:
            if expected_condition_map:
                raise exception.BackendInteractionException(
                    "Specifying expected_condition_map and"
                    "if_not_exist is not allowed both"
                )
            if self._put_item_if_not_exists(table_info,
                                            put_request.attribute_map):
                return True
            raise ConditionalCheckFailedException()
        elif table_info.schema.index_def_map:
            while True:
                old_indexes = self._select_current_index_values(
                    table_info, put_request.attribute_map
                )

                if old_indexes is None:
                    if self._put_item_if_not_exists(table_info,
                                                    put_request.attribute_map):
                        return True
                    continue

                query_builder = self._append_update_query(
                    table_info, put_request.attribute_map,
                    expected_condition_map=expected_condition_map, rewrite=True
                )

                if_prefix = " AND " if expected_condition_map else " IF "
                for index_name, index_def in (
                        table_info.schema.index_def_map.iteritems()):
                    old_index_value = old_indexes.get(
                        index_def.attribute_to_index, None
                    )

                    query_builder += (
                        if_prefix, '"', USER_PREFIX,
                        index_def.attribute_to_index, '"=',
                        _encode_predefined_attr_value(old_index_value)
                        if old_index_value else "null"
                    )
                    if_prefix = " AND "

                qb_len = len(query_builder)

                self._append_update_indexes_queries(
                    table_info, old_indexes, put_request.attribute_map,
                    query_builder, rewrite=True
                )
                if len(query_builder) > qb_len:
                    query_builder.insert(0, self._get_batch_begin_clause())
                    query_builder.append(self._get_batch_apply_clause())
                result = self.__cluster_handler.execute_query(
                    "".join(query_builder), consistent=True
                )

                if result[0]['[applied]']:
                    return True

                for attr_name, attr_value in old_indexes.iteritems():
                    cas_name = USER_PREFIX + attr_name
                    (_, current_value) = _decode_predefined_attr(
                        table_info, cas_name, result[0][cas_name])
                    if current_value != attr_value:
                        # index consistency condition wasn't passed
                        break
                else:
                    raise ConditionalCheckFailedException()
        elif expected_condition_map:
            query_builder = self._append_update_query(
                table_info, put_request.attribute_map,
                expected_condition_map=expected_condition_map, rewrite=True
            )
            result = self.__cluster_handler.execute_query(
                "".join(query_builder), consistent=True
            )
            if result[0]['[applied]']:
                return True
            raise ConditionalCheckFailedException()
        else:
            query_builder = self._append_insert_query(
                table_info, put_request.attribute_map
            )
            self.__cluster_handler.execute_query("".join(query_builder),
                                                 consistent=True)
            return True

    @classmethod
    def _append_delete_query_with_basic_pk(
            cls, table_info, attribute_map, query_builder=None):
        if query_builder is None:
            query_builder = []
        query_builder += (
            'DELETE FROM "', table_info.internal_keyspace, '"."',
            table_info.internal_name, '"'
        )
        cls._append_primary_key(table_info.schema, attribute_map,
                                query_builder)
        return query_builder

    @classmethod
    def _append_delete_query(
            cls, table_info, attribute_map, query_builder=None,
            index_name=None, index_value=None, expected_condition_map=None):
        query_builder = cls._append_delete_query_with_basic_pk(
            table_info, attribute_map, query_builder)

        if table_info.schema.index_def_map:
            cls._append_index_extra_primary_key(query_builder, index_name,
                                                index_value)

        if expected_condition_map:
            cls._append_expected_conditions(
                expected_condition_map, table_info.schema,
                query_builder
            )

        return query_builder

    def _select_current_index_values(
            self, table_info, attribute_map):
        query_buider = ["SELECT "]
        prefix = ""
        for index_def in table_info.schema.index_def_map.itervalues():
            query_buider += (
                prefix, '"', USER_PREFIX, index_def.attribute_to_index, '"'
            )
            prefix = ","

        query_buider += (
            ' FROM "', table_info.internal_keyspace, '"."',
            table_info.internal_name, '"'
        )

        self._append_primary_key(table_info.schema, attribute_map,
                                 query_buider)
        self._append_index_extra_primary_key(query_buider, prefix=" AND ")

        select_result = self.__cluster_handler.execute_query(
            "".join(query_buider), consistent=False
        )
        if not select_result:
            return None

        assert len(select_result) == 1
        index_values = {}

        for cas_attr_name, cas_attr_value in select_result[0].iteritems():
            if cas_attr_value:
                attr_name = cas_attr_name[USER_PREFIX_LENGTH:]
                attr_type = table_info.schema.attribute_type_map[
                    attr_name
                ]
                index_values[attr_name] = (
                    models.AttributeValue(attr_type, cas_attr_value)
                )
        return index_values

    def delete_item(self, context, delete_request,
                    expected_condition_map=None):
        """
        @param context: current request context
        @param delete_request: contains DeleteItemRequest items to perform
                    delete item operation
        @param expected_condition_map: expected attribute name to
                    ExpectedCondition instance mapping. It provides
                    preconditions to make decision about should item be deleted
                    or not

        @return: True if operation performed, otherwise False (if operation was
                    skipped by out of date timestamp, it is considered as
                    successfully performed)

        @raise BackendInteractionException
        """

        table_info = self.__table_info_repo.get(context,
                                                delete_request.table_name)
        delete_query = "".join(
            self._append_delete_query(
                table_info, delete_request.key_attribute_map,
                expected_condition_map=expected_condition_map
            )
        )

        if table_info.schema.index_def_map:
            while True:
                old_indexes = self._select_current_index_values(
                    table_info, delete_request.key_attribute_map
                )

                if old_indexes is None:
                    # Nothing to delete
                    if expected_condition_map:
                        raise ConditionalCheckFailedException()
                    return True

                query_builder = [delete_query]
                if_prefix = " AND " if expected_condition_map else " IF "
                for index_name, index_def in (
                        table_info.schema.index_def_map.iteritems()):
                    index_value = old_indexes.get(
                        index_def.attribute_to_index, None
                    )
                    query_builder += (
                        if_prefix, '"', USER_PREFIX,
                        index_def.attribute_to_index, '"=',
                        _encode_predefined_attr_value(index_value)
                        if index_value
                        else "null"
                    )
                    if_prefix = " AND "

                qb_len = len(query_builder)

                self._append_update_indexes_queries(
                    table_info, old_indexes, delete_request.key_attribute_map,
                    query_builder
                )
                if len(query_builder) > qb_len:
                    query_builder.insert(0, self._get_batch_begin_clause())
                    query_builder.append(self._get_batch_apply_clause())
                result = self.__cluster_handler.execute_query(
                    "".join(query_builder), consistent=True
                )

                if result[0]['[applied]']:
                    return True

                for attr_name, attr_value in old_indexes.iteritems():
                    cas_name = USER_PREFIX + attr_name
                    (_, current_value) = _decode_predefined_attr(
                        table_info, cas_name, result[0][cas_name])
                    if current_value != attr_value:
                        # index consistency condition wasn't passed
                        break
                else:
                    # expected condition wasn't passed
                    raise ConditionalCheckFailedException()
        else:
            result = self.__cluster_handler.execute_query(delete_query,
                                                          consistent=True)
            if result and not result[0]['[applied]']:
                raise ConditionalCheckFailedException()
            return True

    @staticmethod
    def _compact_indexed_condition(cond_list):
        left_condition = None
        right_condition = None
        exact_condition = None

        assert cond_list

        for condition in cond_list:
            if condition.type == models.IndexedCondition.CONDITION_TYPE_EQUAL:
                if (exact_condition is not None and
                        condition.arg.value != exact_condition.arg.value):
                    return None
                exact_condition = condition
            elif condition.is_left_border():
                if left_condition is None:
                    left_condition = condition
                elif condition.is_strict_border():
                    if condition.arg.value >= left_condition.arg.value:
                        left_condition = condition
                else:
                    if condition.arg.value > left_condition.arg.value:
                        left_condition = condition
            elif condition.is_right_border():
                if right_condition is None:
                    right_condition = condition
                elif condition.is_strict():
                    if condition.arg.value <= right_condition.arg.value:
                        right_condition = condition
                else:
                    if condition.arg.value < right_condition.arg.value:
                        right_condition = condition

        if exact_condition is not None:
            if left_condition is not None:
                if left_condition.is_strict():
                    if left_condition.arg.value >= exact_condition.arg.value:
                        return None
                else:
                    if left_condition.arg.value > exact_condition.arg.value:
                        return None
            if right_condition is not None:
                if right_condition.is_strict():
                    if right_condition.arg.value <= exact_condition.arg.value:
                        return None
                else:
                    if right_condition.arg.value < exact_condition.arg.value:
                        return None
            return [exact_condition]
        elif left_condition is not None:
            if right_condition is not None:
                if (left_condition.is_strict_border() or
                        right_condition.is_strict_border()):
                    if left_condition.arg.value >= right_condition.arg.value:
                        return None
                else:
                    if left_condition.arg.value > right_condition.arg.value:
                        return None
                return [left_condition, right_condition]
            else:
                return [left_condition]

        assert right_condition is not None

        return [right_condition]

    def _append_indexed_condition(self, attr_name, condition, query_builder,
                                  column_prefix=USER_PREFIX):
        if query_builder is None:
            query_builder = []
        op = CONDITION_TO_OP[condition.type]
        query_builder += (
            '"', column_prefix, attr_name, '"', op,
            _encode_predefined_attr_value(condition.arg)
        )
        return query_builder

    def _append_hash_key_indexed_condition(
            self, attr_name, condition, query_builder,
            column_prefix=USER_PREFIX):
        if condition.type == models.IndexedCondition.CONDITION_TYPE_EQUAL:
            return self._append_indexed_condition(
                attr_name, condition, query_builder, column_prefix
            )
        else:
            op = CONDITION_TO_OP[condition.type]
            if query_builder is None:
                query_builder = []
            query_builder += (
                'token("', column_prefix, attr_name, '")', op, "token(",
                _encode_predefined_attr_value(condition.arg), ")"
            )
            return query_builder

    @classmethod
    def _append_expected_conditions(cls, expected_condition_map, schema,
                                    query_builder, prefix=" IF "):
        if query_builder is None:
            query_builder = []
        for attr_name, cond_list in expected_condition_map.iteritems():
            for condition in cond_list:
                query_builder.append(prefix)
                cls._append_expected_condition(
                    attr_name, condition, query_builder,
                    attr_name in schema.attribute_type_map
                )
                prefix = " AND "
        return query_builder

    @staticmethod
    def _append_expected_condition(attr, condition, query_builder,
                                   is_predefined):
        if query_builder is None:
            query_builder = []
        if condition.type == models.ExpectedCondition.CONDITION_TYPE_EXISTS:
            if condition.arg:
                query_builder += (
                    SYSTEM_COLUMN_ATTR_EXIST, "={'", attr, "'}"
                )
            else:
                if is_predefined:
                    query_builder += (
                        '"', USER_PREFIX, attr, '"=null'
                    )
                else:
                    query_builder += (
                        SYSTEM_COLUMN_EXTRA_ATTR_DATA, "['", attr, "']=null"
                    )
        elif condition.type == models.ExpectedCondition.CONDITION_TYPE_EQUAL:
            if is_predefined:
                query_builder += (
                    '"', USER_PREFIX, attr, '"=',
                    _encode_predefined_attr_value(condition.arg)
                )
            else:
                query_builder += (
                    SYSTEM_COLUMN_EXTRA_ATTR_DATA, "['", attr, "']=",
                    _encode_dynamic_attr_value(condition.arg)
                )
        else:
            assert False
        return query_builder

    @staticmethod
    def _append_primary_key(table_schema, attribute_map, query_builder,
                            prefix=" WHERE "):
        if query_builder is None:
            query_builder = []
        for key_attr in table_schema.key_attributes:
            if key_attr:
                query_builder += (
                    prefix, '"', USER_PREFIX, key_attr, '"=',
                    _encode_predefined_attr_value(attribute_map[key_attr])
                )
                prefix = " AND "
        return query_builder

    @staticmethod
    def _append_index_extra_primary_key(query_builder=None,
                                        index_name=None, index_value=None,
                                        prefix=" AND "):
        if query_builder is None:
            query_builder = []

        res_index_values = [
            ENCODED_DEFAULT_STRING_VALUE,
            ENCODED_DEFAULT_STRING_VALUE,
            ENCODED_DEFAULT_NUMBER_VALUE,
            ENCODED_DEFAULT_BLOB_VALUE
        ]

        if index_name:
            res_index_values[0] = _encode_single_value_as_predefined_attr(
                index_name, models.AttributeType.ELEMENT_TYPE_STRING
            )

            res_index_values[INDEX_TYPE_TO_INDEX_POS_MAP[index_value.type]] = (
                _encode_predefined_attr_value(index_value)
            )

        for i in xrange(len(LOCAL_INDEX_FIELD_LIST)):
            query_builder += (
                prefix, LOCAL_INDEX_FIELD_LIST[i], "=", res_index_values[i]
            )
            prefix = " AND "

        return query_builder

    @staticmethod
    def _get_batch_begin_clause(durable=True):
        if durable:
            return 'BEGIN BATCH '
        return 'BEGIN UNLOGGED BATCH '

    @staticmethod
    def _get_batch_apply_clause():
        return ' APPLY BATCH'

    def execute_write_batch(self, context, write_request_list):
        """
        @param context: current request context
        @param write_request_list: contains WriteItemBatchableRequest items to
                    perform batch

        @return: List of unprocessed items

        @raise BackendInteractionException
        """

        assert write_request_list
        unprocessed_items = []
        for req in write_request_list:
            try:
                if isinstance(req, models.PutItemRequest):
                    self.put_item(context, req)
                elif isinstance(req, models.DeleteItemRequest):
                    self.delete_item(context, req)
                else:
                    assert False, 'Wrong WriteItemRequest'
            except exception.BackendInteractionException:
                unprocessed_items.append(req)
                LOG.exception("Can't process WriteItemRequest")

        return unprocessed_items

    def update_item(self, context, table_name, key_attribute_map,
                    attribute_action_map, expected_condition_map=None):
        """
        @param context: current request context
        @param table_name: String, name of table to delete item from
        @param key_attribute_map: key attribute name to
                    AttributeValue mapping. It defines row it to update item
        @param attribute_action_map: attribute name to UpdateItemAction
                    instance mapping. It defines actions to perform for each
                    given attribute
        @param expected_condition_map: expected attribute name to
                    ExpectedCondition instance mapping. It provides
                    preconditions to make decision about should item be updated
                    or not
        @return: True if operation performed, otherwise False

        @raise BackendInteractionException
        """
        attribute_action_map = attribute_action_map or {}

        table_info = self.__table_info_repo.get(context, table_name)

        if table_info.schema.index_def_map:
            index_actions = {}
            for index_name, index_def in (
                    table_info.schema.index_def_map.iteritems()):
                attr_name = index_def.attribute_to_index
                action = attribute_action_map.get(
                    attr_name, None
                )
                if action:
                    index_actions[attr_name] = action

            while True:
                old_indexes = self._select_current_index_values(
                    table_info, key_attribute_map
                )

                if old_indexes is None:
                    if expected_condition_map:
                        raise ConditionalCheckFailedException()

                    attribute_map = key_attribute_map.copy()
                    for attr_name, attr_action in (
                            attribute_action_map.iteritems()):
                        if attr_action.action in (
                                models.UpdateItemAction.UPDATE_ACTION_PUT,
                                models.UpdateItemAction.UPDATE_ACTION_ADD):
                            attribute_map[attr_name] = attr_action.value
                    if self._put_item_if_not_exists(table_info,
                                                    attribute_map):
                        return True
                    continue

                attribute_map = key_attribute_map.copy()
                for attr_name, attr_action in (
                        attribute_action_map.iteritems()):
                    if attr_action.action in (
                            models.UpdateItemAction.UPDATE_ACTION_PUT,
                            models.UpdateItemAction.UPDATE_ACTION_ADD):
                        attribute_map[attr_name] = attr_action.value
                    else:
                        attribute_map[attr_name] = None

                query_builder = self._append_update_query(
                    table_info, attribute_map,
                    expected_condition_map=expected_condition_map
                )

                if_prefix = " AND " if expected_condition_map else " IF "
                for index_name, index_def in (
                        table_info.schema.index_def_map.iteritems()):
                    old_index_value = old_indexes.get(
                        index_def.attribute_to_index, None
                    )
                    query_builder += (
                        if_prefix, '"', USER_PREFIX,
                        index_def.attribute_to_index, '"=',
                        _encode_predefined_attr_value(old_index_value)
                        if old_index_value else "null"
                    )
                    if_prefix = " AND "

                qb_len = len(query_builder)

                self._append_update_indexes_queries(
                    table_info, old_indexes, attribute_map,
                    query_builder
                )
                if len(query_builder) > qb_len:
                    query_builder.insert(0, self._get_batch_begin_clause())
                    query_builder.append(self._get_batch_apply_clause())
                result = self.__cluster_handler.execute_query(
                    "".join(query_builder), consistent=True
                )

                if result[0]['[applied]']:
                    return True

                for attr_name, attr_value in old_indexes.iteritems():
                    cas_name = USER_PREFIX + attr_name
                    (_, current_value) = _decode_predefined_attr(
                        table_info, cas_name, result[0][cas_name])
                    if current_value != attr_value:
                        # index consistency condition wasn't passed
                        break
                else:
                    # expected condition wasn't passed
                    raise ConditionalCheckFailedException()
        else:
            attribute_map = key_attribute_map.copy()
            for attr_name, attr_action in (
                    attribute_action_map.iteritems()):
                if attr_action.action in (
                        models.UpdateItemAction.UPDATE_ACTION_PUT,
                        models.UpdateItemAction.UPDATE_ACTION_ADD):
                    attribute_map[attr_name] = attr_action.value
                else:
                    attribute_map[attr_name] = None

            query_builder = self._append_update_query(
                table_info, attribute_map,
                expected_condition_map=expected_condition_map
            )

            result = self.__cluster_handler.execute_query(
                "".join(query_builder), consistent=True
            )

            if result and not result[0]['[applied]']:
                raise ConditionalCheckFailedException()
            return True

    def select_item(self, context, table_name, indexed_condition_map=None,
                    select_type=None, index_name=None, limit=None,
                    exclusive_start_key=None, consistent=True,
                    order_type=None):
        """
        @param context: current request context
        @param table_name: String, name of table to get item from
        @param indexed_condition_map: indexed attribute name to
                    IndexedCondition instance mapping. It defines rows
                    set to be selected
        @param select_type: SelectType instance. It defines with attributes
                    will be returned. If not specified, default will be used:
                        SelectType.all() for query on table and
                        SelectType.all_projected() for query on index
        @param index_name: String, name of index to search with
        @param limit: maximum count of returned values
        @param exclusive_start_key: key attribute names to AttributeValue
                    instance
        @param consistent: define is operation consistent or not (by default it
                    is not consistent)
        @param order_type: defines order of returned rows, if 'None' - default
                    order will be used

        @return SelectResult instance

        @raise BackendInteractionException
        """

        table_info = self.__table_info_repo.get(context, table_name)

        assert (
            not index_name or (
                table_info.schema.index_def_map and
                index_name in table_info.schema.index_def_map
            )
        ), "index_name '{}' isn't specified in the schema".format(
            index_name
        )
        select_type = select_type or models.SelectType.all()

        query_builder = [
            "SELECT ", 'COUNT(*)' if select_type.is_count else '*', ' FROM "',
            USER_PREFIX, context.tenant, '"."', table_info.internal_name, '"'
        ]

        hash_name = table_info.schema.key_attributes[0]

        range_name = (
            table_info.schema.key_attributes[1]
            if len(table_info.schema.key_attributes) > 1
            else None
        )

        indexed_attr_name = table_info.schema.index_def_map[
            index_name
        ].attribute_to_index if index_name else None

        hash_key_cond_list = []
        index_attr_cond_list = []
        range_condition_list = []

        if indexed_condition_map:
            indexed_condition_map_copy = indexed_condition_map.copy()
            # Extracting conditions
            if hash_name in indexed_condition_map_copy:
                hash_key_cond_list = indexed_condition_map_copy.pop(hash_name)

            if index_name and (
                    indexed_attr_name in indexed_condition_map_copy):
                index_attr_cond_list = indexed_condition_map_copy.pop(
                    indexed_attr_name
                )
            if range_name and range_name in indexed_condition_map_copy:
                range_condition_list = indexed_condition_map_copy.pop(
                    range_name
                )
            assert not indexed_condition_map_copy

        # processing exclusive_start_key and append conditions
        if exclusive_start_key:
            exclusive_start_key_copy = exclusive_start_key.copy()
            exclusive_hash_key_value = exclusive_start_key_copy.pop(
                hash_name, None
            )
            if exclusive_hash_key_value:
                hash_key_cond_list.append(
                    models.IndexedCondition.eq(exclusive_hash_key_value)
                    if range_name else
                    models.IndexedCondition.gt(exclusive_hash_key_value)
                )

            if index_name:
                exclusive_indexed_value = exclusive_start_key_copy.pop(
                    indexed_attr_name
                )
                index_attr_cond_list.append(
                    models.IndexedCondition.le(exclusive_indexed_value)
                    if order_type == models.ORDER_TYPE_DESC else
                    models.IndexedCondition.ge(exclusive_indexed_value)
                )

            if range_name:
                exclusive_range_value = exclusive_start_key_copy.pop(
                    range_name
                )

                range_condition_list.append(
                    models.IndexedCondition.lt(exclusive_range_value)
                    if order_type == models.ORDER_TYPE_DESC else
                    models.IndexedCondition.gt(exclusive_range_value)
                )
            assert not exclusive_start_key_copy

        prefix = " WHERE "

        if hash_key_cond_list:
            hash_key_cond_list = self._compact_indexed_condition(
                hash_key_cond_list
            )
            if not hash_key_cond_list:
                return models.SelectResult()
        if range_condition_list:
            range_condition_list = self._compact_indexed_condition(
                range_condition_list
            )
            if not range_condition_list:
                return models.SelectResult()
        if index_attr_cond_list:
            index_attr_cond_list = self._compact_indexed_condition(
                index_attr_cond_list
            )
            if not index_attr_cond_list:
                return models.SelectResult()

        if hash_key_cond_list:
            for cond in hash_key_cond_list:
                query_builder.append(prefix)
                self._append_hash_key_indexed_condition(
                    hash_name, cond, query_builder
                )
                prefix = " AND "

        if table_info.schema.index_def_map:
            # append local secondary index related attrs
            local_indexes_conditions = {
                SYSTEM_COLUMN_INDEX_NAME: [
                    models.IndexedCondition.eq(
                        models.AttributeValue.str(index_name)
                        if index_name else DEFAULT_STRING_VALUE
                    )
                ],
                SYSTEM_COLUMN_INDEX_VALUE_STRING: [],
                SYSTEM_COLUMN_INDEX_VALUE_NUMBER: [],
                SYSTEM_COLUMN_INDEX_VALUE_BLOB: []
            }

            default_index_values = [
                DEFAULT_STRING_VALUE,
                DEFAULT_NUMBER_VALUE,
                DEFAULT_BLOB_VALUE
            ]
            if index_attr_cond_list:
                indexed_attr_type = table_info.schema.attribute_type_map[
                    indexed_attr_name
                ]
                n = INDEX_TYPE_TO_INDEX_POS_MAP[indexed_attr_type]
                for i in xrange(1, n):
                    local_indexes_conditions[
                        LOCAL_INDEX_FIELD_LIST[i]
                    ].append(
                        models.IndexedCondition.eq(
                            default_index_values[i-1]
                        )
                    )
                for index_attr_cond in index_attr_cond_list:
                    local_indexes_conditions[
                        LOCAL_INDEX_FIELD_LIST[n]
                    ].append(index_attr_cond)

                if range_condition_list:
                    for i in xrange(n+1, len(LOCAL_INDEX_FIELD_LIST)):
                        local_indexes_conditions[
                            LOCAL_INDEX_FIELD_LIST[i]
                        ].append(
                            models.IndexedCondition.lt(
                                default_index_values[i-1]
                            )
                            if order_type == models.ORDER_TYPE_DESC else
                            models.IndexedCondition.gt(
                                default_index_values[i-1]
                            )
                        )
            elif range_condition_list:
                for i in xrange(1, len(LOCAL_INDEX_FIELD_LIST)):
                        local_indexes_conditions[
                            LOCAL_INDEX_FIELD_LIST[i]
                        ].append(
                            models.IndexedCondition.eq(
                                default_index_values[i-1]
                            )
                        )

            if local_indexes_conditions:
                for cas_field_name, cond_list in (
                        local_indexes_conditions.iteritems()):
                    for cond in cond_list:
                        query_builder.append(prefix)
                        self._append_indexed_condition(
                            cas_field_name, cond, query_builder,
                            column_prefix=""
                        )
                        prefix = " AND "

        if range_condition_list:
            for cond in range_condition_list:
                query_builder.append(prefix)
                self._append_indexed_condition(
                    range_name, cond, query_builder
                )
                prefix = " AND "

        # add limit
        if limit:
            query_builder += (" LIMIT ", str(limit))

        # add ordering
        if order_type:
            query_builder.append(' ORDER BY ')
            if table_info.schema.index_def_map:
                query_builder += (
                    SYSTEM_COLUMN_INDEX_NAME, " ", order_type
                )
            elif range_name:
                query_builder += (
                    '"', USER_PREFIX, range_name, '" ', order_type
                )
            else:
                assert False

        if not hash_key_cond_list or (
                hash_key_cond_list[0].type !=
                models.IndexedCondition.CONDITION_TYPE_EQUAL):
            query_builder.append(" ALLOW FILTERING")

        rows = self.__cluster_handler.execute_query(
            "".join(query_builder), consistent
        )

        if select_type.is_count:
            return models.SelectResult(count=rows[0]['count'])

        # process results

        result = []

        # TODO ikhudoshyn: if select_type.is_all_projected,
        # get list of projected attrs by index_name from metainfo

        attributes_to_get = select_type.attributes

        for row in rows:
            record = {}

            # add predefined attributes
            for cas_name, cas_val in row.iteritems():
                if cas_name.startswith(USER_PREFIX) and cas_val:
                    name, val = _decode_predefined_attr(table_info, cas_name,
                                                        cas_val)
                    if not attributes_to_get or name in attributes_to_get:
                        record[name] = val

            # add dynamic attributes (from SYSTEM_COLUMN_ATTR_DATA dict)
            types = row[SYSTEM_COLUMN_EXTRA_ATTR_TYPES]
            attrs = row[SYSTEM_COLUMN_EXTRA_ATTR_DATA] or {}
            for name, val in attrs.iteritems():
                if not attributes_to_get or name in attributes_to_get:
                    typ = types[name]
                    storage_type = CASSANDRA_TO_STORAGE_TYPES[typ]
                    record[name] = _decode_dynamic_value(val, storage_type)

            result.append(record)

        count = len(result)
        if limit and count == limit:
            last_evaluated_key = {hash_name: result[-1][hash_name]}

            if range_name:
                last_evaluated_key[range_name] = result[-1][range_name]

            if index_name:
                indexed_attr_name = table_info.schema.index_def_map[
                    index_name
                ].attribute_to_index
                last_evaluated_key[indexed_attr_name] = result[-1][
                    indexed_attr_name
                ]
        else:
            last_evaluated_key = None

        return models.SelectResult(items=result,
                                   last_evaluated_key=last_evaluated_key,
                                   count=count)

    def scan(self, context, table_name, condition_map, attributes_to_get=None,
             limit=None, exclusive_start_key=None, consistent=False):
        """
        @param context: current request context
        @param table_name: String, name of table to get item from
        @param condition_map: indexed attribute name to
                    IndexedCondition instance mapping. It defines rows
                    set to be selected
        @param limit: maximum count of returned values
        @param exclusive_start_key: key attribute names to AttributeValue
                    instance
        @param consistent: define is operation consistent or not (by default it
                    is not consistent)

        @return list of attribute name to AttributeValue mappings

        @raise BackendInteractionException
        """
        if not condition_map:
            condition_map = {}
        table_info = self.__table_info_repo.get(context, table_name)
        hash_name = table_info.schema.key_attributes[0]
        try:
            range_name = table_info.schema.key_attributes[1]
        except IndexError:
            range_name = None

        key_conditions = {
            hash_name: []
        }

        if range_name:
            key_conditions[range_name] = []

        if hash_name in condition_map:
            key_conditions[hash_name] = condition_map[hash_name]

            if (range_name and range_name in condition_map
                and condition_map[range_name].type in
                    models.IndexedCondition._allowed_types):

                key_conditions[range_name] = condition_map[range_name]
        if exclusive_start_key:
            if range_name:
                key_conditions[hash_name].append(
                    models.IndexedCondition.eq(exclusive_start_key[hash_name])
                )
                key_conditions[range_name].append(
                    models.IndexedCondition.gt(exclusive_start_key[range_name])
                )
            else:
                key_conditions[hash_name].append(
                    models.IndexedCondition.gt(exclusive_start_key[hash_name])
                )

        selected = self.select_item(context, table_name, key_conditions,
                                    models.SelectType.all(), limit=limit,
                                    consistent=consistent)

        if (range_name and exclusive_start_key
                and range_name in exclusive_start_key
                and (not limit or limit > selected.count)):

            del key_conditions[range_name][-1]
            del key_conditions[hash_name][-1]
            key_conditions[hash_name].append(
                models.IndexedCondition.gt(exclusive_start_key[hash_name])
            )

            limit2 = limit - selected.count if limit else None

            selected2 = self.select_item(
                context, table_name, key_conditions,
                models.SelectType.all(), limit=limit2,
                consistent=consistent)

            selected = models.SelectResult(
                items=selected.items + selected2.items,
                last_evaluated_key=selected2.last_evaluated_key,
                count=selected.count + selected2.count
            )

        scanned_count = selected.count

        if selected.items:
            filtered_items = filter(
                lambda item: self._conditions_satisfied(
                    item, condition_map),
                selected.items)
            count = len(filtered_items)
        else:
            filtered_items = []
            count = selected.count

        if attributes_to_get and filtered_items:
            for item in filtered_items:
                for attr in item.keys():
                    if attr not in attributes_to_get:
                        del item[attr]

        filtered = models.ScanResult(
            items=filtered_items,
            last_evaluated_key=selected.last_evaluated_key,
            count=count, scanned_count=scanned_count)

        return filtered

    def _conditions_satisfied(self, row, cond_map=None):
        if not cond_map:
            return True

        for attr_name, cond_list in cond_map.iteritems():
            for cond in cond_list:
                if not self._condition_satisfied(
                        row.get(attr_name, None), cond):
                    return False
        return True

    @staticmethod
    def _condition_satisfied(attr_val, cond):

        if cond.type == models.ExpectedCondition.CONDITION_TYPE_EXISTS:
            return cond.arg == bool(attr_val)

        if not attr_val:
            return False

        if cond.type == models.Condition.CONDITION_TYPE_EQUAL:
            return (attr_val.type == cond.arg.type and
                    attr_val.value == cond.arg.value)

        if cond.type == models.IndexedCondition.CONDITION_TYPE_LESS:
            return (attr_val.type == cond.arg.type and
                    attr_val.value < cond.arg.value)

        if cond.type == models.IndexedCondition.CONDITION_TYPE_LESS_OR_EQUAL:
            return (attr_val.type == cond.arg.type and
                    attr_val.value <= cond.arg.value)

        if cond.type == models.IndexedCondition.CONDITION_TYPE_GREATER:
            return (attr_val.type == cond.arg.type and
                    attr_val.value > cond.arg.value)

        if (cond.type ==
                models.IndexedCondition.CONDITION_TYPE_GREATER_OR_EQUAL):
            return (attr_val.type == cond.arg.type and
                    attr_val.value >= cond.arg.value)

        if cond.type == models.ScanCondition.CONDITION_TYPE_NOT_EQUAL:
            return (attr_val.type != cond.arg.type or
                    attr_val.value != cond.arg.value)

        if cond.type == models.ScanCondition.CONDITION_TYPE_CONTAINS:
            assert not cond.arg.type.collection_type
            if attr_val.type.element_type != cond.arg.type.element_type:
                return False

            return cond.arg.value in attr_val.value

        if cond.type == models.ScanCondition.CONDITION_TYPE_NOT_CONTAINS:
            assert not cond.arg.type.collection_type
            if attr_val.type.element_type != cond.arg.type.element_type:
                return False

            return cond.arg.value not in attr_val.value

        if cond.type == models.ScanCondition.CONDITION_TYPE_IN:
            cond_arg = cond.arg or []

            return attr_val in cond_arg

        return False



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import unittest

from concurrent.futures import Future

from magnetodb.storage import models
from magnetodb.storage.manager.simple_impl import SimpleStorageManager


class SimpleStorageManagerTestCase(unittest.TestCase):
    """The test for simple storage manager implementation."""

    @mock.patch('magnetodb.storage.manager.simple_impl.SimpleStorageManager.'
                'delete_item_async')
    @mock.patch('magnetodb.storage.manager.simple_impl.SimpleStorageManager.'
                'put_item_async')
    def test_execute_write_batch(self, mock_put_item, mock_delete_item):
        future = Future()
        future.set_result(True)
        mock_put_item.return_value = future
        mock_delete_item.return_value = future

        context = mock.Mock(tenant='fake_tenant')

        table_name = 'fake_table'

        request_list = [
            models.PutItemRequest(table_name, {
                'id': models.AttributeValue(
                    models.ATTRIBUTE_TYPE_NUMBER, 1),
                'range': models.AttributeValue(
                    models.ATTRIBUTE_TYPE_STRING, '1'),
                'str': models.AttributeValue(
                    models.ATTRIBUTE_TYPE_STRING, 'str1'), }),
            models.PutItemRequest(table_name, {
                'id': models.AttributeValue(
                    models.ATTRIBUTE_TYPE_NUMBER, 2),
                'range': models.AttributeValue(
                    models.ATTRIBUTE_TYPE_STRING, '1'),
                'str': models.AttributeValue(
                    models.ATTRIBUTE_TYPE_STRING, 'str1'), }),
            models.DeleteItemRequest(table_name, {
                'id': models.AttributeValue.number(3),
                'range': models.AttributeValue.str('3')})
        ]

        expected_put = [mock.call(context, request_list[0]),
                        mock.call(context, request_list[1]), ]
        expected_delete = [mock.call(context, request_list[2])]

        storage_manager = SimpleStorageManager(None, None)

        unprocessed_items = storage_manager.execute_write_batch(
            context, request_list
        )

        self.assertEqual(expected_put, mock_put_item.call_args_list)
        self.assertEqual(expected_delete,
                         mock_delete_item.call_args_list)

        self.assertEqual(unprocessed_items, [])



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import atexit
import functools
import os
import time

import fixtures
import nose.plugins.attrib
import testresources
import testtools

from tempest import clients
# from tempest.common import isolated_creds
from tempest import config
from tempest import exceptions
from tempest.openstack.common import log as logging
from unittest.case import _AssertRaisesContext

LOG = logging.getLogger(__name__)

# All the successful HTTP status codes from RFC 2616
HTTP_SUCCESS = (200, 201, 202, 203, 204, 205, 206)


def attr(*args, **kwargs):
    """A decorator which applies the nose and testtools attr decorator

    This decorator applies the nose attr decorator as well as the
    the testtools.testcase.attr if it is in the list of attributes
    to testtools we want to apply.
    """

    def decorator(f):
        if 'type' in kwargs and isinstance(kwargs['type'], str):
            f = testtools.testcase.attr(kwargs['type'])(f)
            if kwargs['type'] == 'smoke':
                f = testtools.testcase.attr('gate')(f)
        elif 'type' in kwargs and isinstance(kwargs['type'], list):
            for attr in kwargs['type']:
                f = testtools.testcase.attr(attr)(f)
                if attr == 'smoke':
                    f = testtools.testcase.attr('gate')(f)
        return nose.plugins.attrib.attr(*args, **kwargs)(f)

    return decorator


def services(*args, **kwargs):
    """A decorator used to set an attr for each service used in a test case

    This decorator applies a testtools attr for each service that gets
    exercised by a test case.
    """
    valid_service_list = ['compute', 'image', 'volume', 'orchestration',
                          'network', 'identity', 'object', 'dashboard']

    def decorator(f):
        for service in args:
            if service not in valid_service_list:
                raise exceptions.InvalidServiceTag('%s is not a valid service'
                                                   % service)
        attr(type=list(args))(f)
        return f
    return decorator


def stresstest(*args, **kwargs):
    """Add stress test decorator

    For all functions with this decorator a attr stress will be
    set automatically.

    @param class_setup_per: allowed values are application, process, action
           ``application``: once in the stress job lifetime
           ``process``: once in the worker process lifetime
           ``action``: on each action
    @param allow_inheritance: allows inheritance of this attribute
    """
    def decorator(f):
        if 'class_setup_per' in kwargs:
            setattr(f, "st_class_setup_per", kwargs['class_setup_per'])
        else:
            setattr(f, "st_class_setup_per", 'process')
        if 'allow_inheritance' in kwargs:
            setattr(f, "st_allow_inheritance", kwargs['allow_inheritance'])
        else:
            setattr(f, "st_allow_inheritance", False)
        attr(type='stress')(f)
        return f
    return decorator


def skip_because(*args, **kwargs):
    """A decorator useful to skip tests hitting known bugs

    @param bug: bug number causing the test to skip
    @param condition: optional condition to be True for the skip to have place
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*func_args, **func_kwargs):
            if "bug" in kwargs:
                if "condition" not in kwargs or kwargs["condition"] is True:
                    msg = "Skipped until Bug: %s is resolved." % kwargs["bug"]
                    raise testtools.TestCase.skipException(msg)
            return f(*func_args, **func_kwargs)
        return wrapper
    return decorator


# there is a mis-match between nose and testtools for older pythons.
# testtools will set skipException to be either
# unittest.case.SkipTest, unittest2.case.SkipTest or an internal skip
# exception, depending on what it can find. Python <2.7 doesn't have
# unittest.case.SkipTest; so if unittest2 is not installed it falls
# back to the internal class.
#
# The current nose skip plugin will decide to raise either
# unittest.case.SkipTest or its own internal exception; it does not
# look for unittest2 or the internal unittest exception.  Thus we must
# monkey-patch testtools.TestCase.skipException to be the exception
# the nose skip plugin expects.
#
# However, with the switch to testr nose may not be available, so we
# require you to opt-in to this fix with an environment variable.
#
# This is temporary until upstream nose starts looking for unittest2
# as testtools does; we can then remove this and ensure unittest2 is
# available for older pythons; then nose and testtools will agree
# unittest2.case.SkipTest is the one-true skip test exception.
#
#   https://review.openstack.org/#/c/33056
#   https://github.com/nose-devs/nose/pull/699
if 'TEMPEST_PY26_NOSE_COMPAT' in os.environ:
    try:
        import unittest.case.SkipTest
        # convince pep8 we're using the import...
        if unittest.case.SkipTest:
            pass
        raise RuntimeError("You have unittest.case.SkipTest; "
                           "no need to override")
    except ImportError:
        LOG.info("Overriding skipException to nose SkipTest")
        testtools.TestCase.skipException = nose.plugins.skip.SkipTest

at_exit_set = set()


def validate_tearDownClass():
    if at_exit_set:
        raise RuntimeError("tearDownClass does not calls the super's "
                           "tearDownClass in these classes: "
                           + str(at_exit_set) + "\n"
                           "If you see the exception, with another "
                           "exception please do not report this one!"
                           "If you are changing tempest code, make sure you",
                           "are calling the super class's tearDownClass!")

atexit.register(validate_tearDownClass)


class BaseTestCase(testtools.TestCase,
                   testtools.testcase.WithAttributes,
                   testresources.ResourcedTestCase):

    config = config.TempestConfig()

    setUpClassCalled = False

    @classmethod
    def setUpClass(cls):
        if hasattr(super(BaseTestCase, cls), 'setUpClass'):
            super(BaseTestCase, cls).setUpClass()
        cls.setUpClassCalled = True

    @classmethod
    def tearDownClass(cls):
        at_exit_set.discard(cls)
        if hasattr(super(BaseTestCase, cls), 'tearDownClass'):
            super(BaseTestCase, cls).tearDownClass()

    def setUp(self):
        super(BaseTestCase, self).setUp()
        if not self.setUpClassCalled:
            raise RuntimeError("setUpClass does not calls the super's"
                               "setUpClass in the "
                               + self.__class__.__name__)
        at_exit_set.add(self.__class__)
        test_timeout = os.environ.get('OS_TEST_TIMEOUT', 0)
        try:
            test_timeout = int(test_timeout)
        except ValueError:
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

        if (os.environ.get('OS_STDOUT_CAPTURE') == 'True' or
                os.environ.get('OS_STDOUT_CAPTURE') == '1'):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if (os.environ.get('OS_STDERR_CAPTURE') == 'True' or
                os.environ.get('OS_STDERR_CAPTURE') == '1'):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))
        if (os.environ.get('OS_LOG_CAPTURE') != 'False' and
                os.environ.get('OS_LOG_CAPTURE') != '0'):
            log_format = '%(asctime)-15s %(message)s'
            self.useFixture(fixtures.LoggerFixture(nuke_handlers=False,
                                                   format=log_format,
                                                   level=None))

    @classmethod
    def get_client_manager(cls):
        """
        Returns an Openstack client manager
        """
        # cls.isolated_creds = isolated_creds.IsolatedCreds(cls.__name__)

        force_tenant_isolation = getattr(cls, 'force_tenant_isolation', None)
        if (cls.config.compute.allow_tenant_isolation or
                force_tenant_isolation):
            # TODO(yyekovenko) temp. to avoid usage of network client
            # creds = cls.isolated_creds.get_primary_creds()
            # username, tenant_name, password = creds
            # os = clients.Manager(username=username,
            #                     password=password,
            #                     tenant_name=tenant_name,
            #                     interface=cls._interface)
            os = clients.Manager(username=None,
                                 password=None,
                                 tenant_name=None,
                                 interface=cls._interface)
        else:
            os = clients.Manager(interface=cls._interface)
        return os

    # @classmethod
    # def clear_isolated_creds(cls):
    #    """
    #    Clears isolated creds if set
    #    """
    #    if getattr(cls, 'isolated_creds'):
    #        cls.isolated_creds.clear_isolated_creds()

    @classmethod
    def _get_identity_admin_client(cls):
        """
        Returns an instance of the Identity Admin API client
        """
        os = clients.AdminManager(interface=cls._interface)
        admin_client = os.identity_client
        return admin_client

    @classmethod
    def _get_client_args(cls):

        return (
            cls.config,
            cls.config.identity.admin_username,
            cls.config.identity.admin_password,
            cls.config.identity.uri
        )

    def assertRaises(self, excClass, callableObj=None, *args, **kwargs):
        context = _AssertRaisesContext(excClass, self)
        if callableObj is None:
            return context
        with context:
            callableObj(*args, **kwargs)


def call_until_true(func, duration, sleep_for):
    """
    Call the given function until it returns True (and return True) or
    until the specified duration (in seconds) elapses (and return
    False).

    :param func: A zero argument callable that returns True on success.
    :param duration: The number of seconds for which to attempt a
        successful call of the function.
    :param sleep_for: The number of seconds to sleep after an unsuccessful
                      invocation of the function.
    """
    now = time.time()
    timeout = now + duration
    while now < timeout:
        if func():
            return True
        LOG.debug("Sleeping for %d seconds", sleep_for)
        time.sleep(sleep_for)
        now = time.time()
    return False



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.api.keyvalue.boto_base.base import MagnetoDBTestCase
from tempest.common.utils.data_utils import rand_name


class MagnetoDBQueriesTest(MagnetoDBTestCase):

    @classmethod
    def setUpClass(cls):
        super(MagnetoDBQueriesTest, cls).setUpClass()
        cls.tname = rand_name().replace('-', '')
        cls.client.create_table(cls.smoke_attrs,
                                cls.tname,
                                cls.smoke_schema,
                                cls.smoke_throughput)
        cls.wait_for_table_active(cls.tname)
        cls.addResourceCleanUp(cls.client.delete_table, cls.tname)



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import jsonschema

from magnetodb.api.openstack.v1 import parser
from magnetodb.api.openstack.v1 import utils
from magnetodb import storage
from magnetodb.storage import models


class GetItemController(object):
    schema = {
        "required": [parser.Props.KEY],
        "properties": {
            parser.Props.ATTRIBUTES_TO_GET: {
                "type": "array",
                "items": {
                    "type": "string",
                    "pattern": parser.ATTRIBUTE_NAME_PATTERN
                }
            },
            parser.Props.CONSISTENT_READ: {
                "type": "boolean"
            },
            parser.Props.KEY: {
                "type": "object",
                "patternProperties": {
                    parser.ATTRIBUTE_NAME_PATTERN: parser.Types.ITEM_VALUE
                }
            },
        }
    }

    def process_request(self, req, body, project_id, table_name):
        utils.check_project_id(req.context, project_id)
        jsonschema.validate(body, self.schema)
        req.context.tenant = project_id

        # get attributes_to_get
        attributes_to_get = body.get(parser.Props.ATTRIBUTES_TO_GET)

        select_type = (
            models.SelectType.all()
            if attributes_to_get is None else
            models.SelectType.specified_attributes(attributes_to_get)
        )

        # parse key_attributes
        key_attributes = parser.Parser.parse_item_attributes(
            body[parser.Props.KEY]
        )

        # parse consistent_read
        consistent_read = body.get(
            parser.Props.CONSISTENT_READ, False
        )

        # format conditions to get item
        indexed_condition_map = {
            name: [models.IndexedCondition.eq(value)]
            for name, value in key_attributes.iteritems()
        }

        # get item
        result = storage.select_item(
            req.context, table_name, indexed_condition_map,
            select_type=select_type, limit=2, consistent=consistent_read)

        # format response
        if result.count == 0:
            return {}

        response = {
            parser.Props.ITEM: parser.Parser.format_item_attributes(
                result.items[0])
        }
        return response



# Copyright 2013 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

from magnetodb.common import config

from magnetodb.openstack.common import importutils

from magnetodb.openstack.common import log as logging

LOG = logging.getLogger(__name__)

__STORAGE_MANAGER_IMPL = None


def process_object_spec(obj_name, obj_spec_map, context):
    if obj_name in context:
        return context[obj_name]

    obj_spec = obj_spec_map[obj_name]

    if obj_name not in obj_spec_map:
        raise AttributeError(
            "Object with name {} is not found".format(
                obj_name
            )
        )

    params = obj_spec["params"]
    res_params = {}
    for param_name, param_value in params.iteritems():
        if (isinstance(param_value, basestring) and
                len(param_value) > 1 and param_value[0] == "@"):
            param_value = param_value[1:]
            if param_value[0] != "@":
                param_value = process_object_spec(
                    param_value, obj_spec_map, context
                )
        res_params[param_name] = param_value

    type_str = obj_spec["type"]
    mod_str, _sep, class_str = type_str.rpartition('.')
    if mod_str:
        module = importutils.import_module(mod_str)
        cls = getattr(module, class_str)
    else:
        cls = eval(type_str)
    obj = cls(**res_params)
    context[obj_name] = obj

    return obj


def setup():
    global __STORAGE_MANAGER_IMPL
    assert __STORAGE_MANAGER_IMPL is None

    storage_manager_config = config.CONF.storage_manager_config

    object_spec_map = json.loads(storage_manager_config)

    context = {}

    for object_name in object_spec_map.keys():
        process_object_spec(object_name, object_spec_map, context)

    __STORAGE_MANAGER_IMPL = context["storage_manager"]


def create_table(context, table_name, table_schema):
    """
    Creates table

    @param context: current request context
    @param table_name: String, name of the table to create
    @param table_schema: TableSchema instance which define table to create

    @return TableMeta instance with metadata of created table

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.create_table(context, table_name,
                                               table_schema)


def delete_table(context, table_name):
    """
    Delete table

    @param context: current request context
    @param table_name: String, name of table to delete

    @return TableMeta instance with metadata of created table

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.delete_table(context, table_name)


def describe_table(context, table_name):
    """
    Describe table

    @param context: current request context
    @param table_name: String, name of table to describes

    @return: TableMeta instance

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.describe_table(context, table_name)


def list_tables(context, exclusive_start_table_name=None, limit=None):
    """
    @param context: current request context
    @param exclusive_start_table_name
    @param limit: limit of returned table names
    @return list of table names

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.list_tables(
        context, exclusive_start_table_name, limit
    )


def put_item(context, put_request, if_not_exist=False,
             expected_condition_map=None):
    """
    @param context: current request context
    @param put_request: PutItemRequest instance which specifies item to put
    @param if_not_exist: put item only is row is new record (It is possible
                to use only one of if_not_exist and expected_condition_map
                parameter)
    @param expected_condition_map: expected attribute name to
                ExpectedCondition instance mapping. It provides
                preconditions to make decision about should item be put or
                not

    @return: True if operation performed, otherwise False

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.put_item(
        context, put_request, if_not_exist, expected_condition_map
    )


def put_item_async(context, put_request, if_not_exist=False,
                   expected_condition_map=None):
    """
    @param context: current request context
    @param put_request: PutItemRequest instance which specifies item to put
    @param if_not_exist: put item only is row is new record (It is possible
                to use only one of if_not_exist and expected_condition_map
                parameter)
    @param expected_condition_map: expected attribute name to
                ExpectedCondition instance mapping. It provides
                preconditions to make decision about should item be put or
                not

    @return: Future instance

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.put_item_async(
        context, put_request, if_not_exist, expected_condition_map
    )


def delete_item(context, delete_request, expected_condition_map=None):
    """
    @param context: current request context
    @param delete_request: DeleteItemRequest instance which identifies item
                to delete
    @param expected_condition_map: expected attribute name to
                ExpectedCondition instance mapping. It provides
                preconditions to make decision about should item be deleted
                or not

    @return: True if operation performed, otherwise False (if operation was
                skipped by out of date timestamp, it is considered as
                successfully performed)

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.delete_item(
        context, delete_request, expected_condition_map
    )


def delete_item_async(context, delete_request,
                      expected_condition_map=None):
    """
    @param context: current request context
    @param delete_request: DeleteItemRequest instance which identifies item
                to delete
    @param expected_condition_map: expected attribute name to
                ExpectedCondition instance mapping. It provides
                preconditions to make decision about should item be deleted
                or not

    @return: Future instance

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.delete_item_async(
        context, delete_request, expected_condition_map
    )


def execute_write_batch(context, write_request_list):
    """
    @param context: current request context
    @param write_request_list: contains PutItemRequest or DeleteItemRequest
                instances to execute batch operation

    @return: Unprocessed request list
    """
    return __STORAGE_MANAGER_IMPL.execute_write_batch(context,
                                                      write_request_list)


def update_item(context, table_name, key_attribute_map,
                attribute_action_map, expected_condition_map=None):
    """
    @param context: current request context
    @param table_name: String, name of table to delete item from
    @param key_attribute_map: key attribute name to
                AttributeValue mapping. It defines row it to update item
    @param attribute_action_map: attribute name to UpdateItemAction
                instance mapping. It defines actions to perform for each
                given attribute
    @param expected_condition_map: expected attribute name to
                ExpectedCondition instance mapping. It provides
                preconditions
                to make decision about should item be updated or not
    @return: True if operation performed, otherwise False

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.update_item(
        context, table_name, key_attribute_map, attribute_action_map,
        expected_condition_map
    )


def select_item(context, table_name, indexed_condition_map=None,
                select_type=None, index_name=None, limit=None,
                exclusive_start_key=None, consistent=True,
                order_type=None):
    """
    @param context: current request context
    @param table_name: String, name of table to get item from
    @param indexed_condition_map: indexed attribute name to
                IndexedCondition instance mapping. It defines rows
                set to be selected
    @param select_type: SelectType instance. It defines with attributes
                will be returned. If not specified, default will be used:
                    SelectType.all() for query on table and
                    SelectType.all_projected() for query on index
    @param index_name: String, name of index to search with
    @param limit: maximum count of returned values
    @param exclusive_start_key: key attribute names to AttributeValue
                instance
    @param consistent: define is operation consistent or not (by default it
                is not consistent)
    @param order_type: defines order of returned rows, if 'None' - default
                order will be used

    @return SelectResult instance

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.select_item(
        context, table_name, indexed_condition_map, select_type, index_name,
        limit, exclusive_start_key, consistent, order_type
    )


def scan(context, table_name, condition_map, attributes_to_get=None,
         limit=None, exclusive_start_key=None,
         consistent=False):
    """
    @param context: current request context
    @param table_name: String, name of table to get item from
    @param condition_map: attribute name to
                IndexedCondition instance mapping. It defines rows
                set to be selected
    @param attributes_to_get: list of attribute names to be included in
                result. If None, all attributes will be included
    @param limit: maximum count of returned values
    @param exclusive_start_key: key attribute names to AttributeValue
                instance
    @param consistent: define is operation consistent or not (by default it
                is not consistent)

    @return list of attribute name to AttributeValue mappings

    @raise BackendInteractionException
    """
    return __STORAGE_MANAGER_IMPL.scan(
        context, table_name, condition_map, attributes_to_get, limit,
        exclusive_start_key, consistent=False
    )



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function

import os
import sys


from oslo.config import cfg

from tempest.common.utils.misc import singleton
from tempest.openstack.common import log as logging


def register_opt_group(conf, opt_group, options):
    conf.register_group(opt_group)
    for opt in options:
        conf.register_opt(opt, group=opt_group.name)


identity_group = cfg.OptGroup(name='identity',
                              title="Keystone Configuration Options")

IdentityGroup = [
    cfg.StrOpt('catalog_type',
               default='identity',
               help="Catalog type of the Identity service."),
    cfg.BoolOpt('disable_ssl_certificate_validation',
                default=False,
                help="Set to True if using self-signed SSL certificates."),
    cfg.StrOpt('uri',
               default=None,
               help="Full URI of the OpenStack Identity API (Keystone), v2"),
    cfg.StrOpt('uri_v3',
               help='Full URI of the OpenStack Identity API (Keystone), v3'),
    cfg.StrOpt('region',
               default='RegionOne',
               help="The identity region name to use. Also used as the other "
                    "services' region name unless they are set explicitly. "
                    "If no such region is found in the service catalog, the "
                    "first found one is used."),
    cfg.StrOpt('username',
               default='demo',
               help="Username to use for Nova API requests."),
    cfg.StrOpt('tenant_name',
               default='demo',
               help="Tenant name to use for Nova API requests."),
    cfg.StrOpt('admin_role',
               default='admin',
               help="Role required to administrate keystone."),
    cfg.StrOpt('password',
               default='pass',
               help="API key to use when authenticating.",
               secret=True),
    cfg.StrOpt('alt_username',
               default=None,
               help="Username of alternate user to use for Nova API "
                    "requests."),
    cfg.StrOpt('alt_tenant_name',
               default=None,
               help="Alternate user's Tenant name to use for Nova API "
                    "requests."),
    cfg.StrOpt('alt_password',
               default=None,
               help="API key to use when authenticating as alternate user.",
               secret=True),
    cfg.StrOpt('admin_username',
               default='admin',
               help="Administrative Username to use for"
                    "Keystone API requests."),
    cfg.StrOpt('admin_tenant_name',
               default='admin',
               help="Administrative Tenant name to use for Keystone API "
                    "requests."),
    cfg.StrOpt('admin_password',
               default='pass',
               help="API key to use when authenticating as admin.",
               secret=True),
]

compute_group = cfg.OptGroup(name='compute',
                             title='Compute Service Options')

ComputeGroup = [
    cfg.BoolOpt('allow_tenant_isolation',
                default=False,
                help="Allows test cases to create/destroy tenants and "
                     "users. This option enables isolated test cases and "
                     "better parallel execution, but also requires that "
                     "OpenStack Identity API admin credentials are known."),
    cfg.StrOpt('image_ref',
               default="{$IMAGE_ID}",
               help="Valid secondary image reference to be used in tests."),
    cfg.StrOpt('image_ref_alt',
               default="{$IMAGE_ID_ALT}",
               help="Valid secondary image reference to be used in tests."),
    cfg.StrOpt('flavor_ref',
               default="1",
               help="Valid primary flavor to use in tests."),
    cfg.StrOpt('flavor_ref_alt',
               default="2",
               help='Valid secondary flavor to be used in tests.'),
    cfg.StrOpt('image_ssh_user',
               default="root",
               help="User name used to authenticate to an instance."),
    cfg.StrOpt('image_ssh_password',
               default="password",
               help="Password used to authenticate to an instance."),
    cfg.StrOpt('image_alt_ssh_user',
               default="root",
               help="User name used to authenticate to an instance using "
                    "the alternate image."),
    cfg.StrOpt('image_alt_ssh_password',
               default="password",
               help="Password used to authenticate to an instance using "
                    "the alternate image."),
    cfg.IntOpt('build_interval',
               default=10,
               help="Time in seconds between build status checks."),
    cfg.IntOpt('build_timeout',
               default=300,
               help="Timeout in seconds to wait for an instance to build."),
    cfg.BoolOpt('run_ssh',
                default=False,
                help="Does the test environment support snapshots?"),
    cfg.StrOpt('ssh_user',
               default='root',
               help="User name used to authenticate to an instance."),
    cfg.IntOpt('ping_timeout',
               default=60,
               help="Timeout in seconds to wait for ping to "
                    "succeed."),
    cfg.IntOpt('ssh_timeout',
               default=300,
               help="Timeout in seconds to wait for authentication to "
                    "succeed."),
    cfg.IntOpt('ready_wait',
               default=0,
               help="Additional wait time for clean state, when there is "
                    "no OS-EXT-STS extension available"),
    cfg.IntOpt('ssh_channel_timeout',
               default=60,
               help="Timeout in seconds to wait for output from ssh "
                    "channel."),
    cfg.StrOpt('fixed_network_name',
               default='private',
               help="Visible fixed network name "),
    cfg.StrOpt('network_for_ssh',
               default='public',
               help="Network used for SSH connections."),
    cfg.IntOpt('ip_version_for_ssh',
               default=4,
               help="IP version used for SSH connections."),
    cfg.BoolOpt('use_floatingip_for_ssh',
                default=True,
                help="Dose the SSH uses Floating IP?"),
    cfg.StrOpt('catalog_type',
               default='compute',
               help="Catalog type of the Compute service."),
    cfg.StrOpt('region',
               default='',
               help="The compute region name to use. If empty, the value "
                    "of identity.region is used instead. If no such region "
                    "is found in the service catalog, the first found one is "
                    "used."),
    cfg.StrOpt('catalog_v3_type',
               default='computev3',
               help="Catalog type of the Compute v3 service."),
    cfg.StrOpt('path_to_private_key',
               default=None,
               help="Path to a private key file for SSH access to remote "
                    "hosts"),
    cfg.StrOpt('volume_device_name',
               default='vdb',
               help="Expected device name when a volume is attached to "
                    "an instance"),
    cfg.IntOpt('shelved_offload_time',
               default=0,
               help='Time in seconds before a shelved instance is eligible '
                    'for removing from a host.  -1 never offload, 0 offload '
                    'when shelved. This time should be the same as the time '
                    'of nova.conf, and some tests will run for as long as the '
                    'time.')
]

compute_features_group = cfg.OptGroup(name='compute-feature-enabled',
                                      title="Enabled Compute Service Features")

ComputeFeaturesGroup = [
    cfg.BoolOpt('api_v3',
                default=True,
                help="If false, skip all nova v3 tests."),
    cfg.BoolOpt('disk_config',
                default=True,
                help="If false, skip disk config tests"),
    cfg.BoolOpt('flavor_extra',
                default=True,
                help="If false, skip flavor extra data test"),
    cfg.BoolOpt('change_password',
                default=False,
                help="Does the test environment support changing the admin "
                     "password?"),
    cfg.BoolOpt('create_image',
                default=False,
                help="Does the test environment support snapshots?"),
    cfg.BoolOpt('resize',
                default=False,
                help="Does the test environment support resizing?"),
    cfg.BoolOpt('live_migration',
                default=False,
                help="Does the test environment support live migration "
                     "available?"),
    cfg.BoolOpt('block_migration_for_live_migration',
                default=False,
                help="Does the test environment use block devices for live "
                     "migration"),
    cfg.BoolOpt('block_migrate_cinder_iscsi',
                default=False,
                help="Does the test environment block migration support "
                     "cinder iSCSI volumes")
]


compute_admin_group = cfg.OptGroup(name='compute-admin',
                                   title="Compute Admin Options")

ComputeAdminGroup = [
    cfg.StrOpt('username',
               default='admin',
               help="Administrative Username to use for Nova API requests."),
    cfg.StrOpt('tenant_name',
               default='admin',
               help="Administrative Tenant name to use for Nova API "
                    "requests."),
    cfg.StrOpt('password',
               default='pass',
               help="API key to use when authenticating as admin.",
               secret=True),
]

image_group = cfg.OptGroup(name='image',
                           title="Image Service Options")

ImageGroup = [
    cfg.StrOpt('catalog_type',
               default='image',
               help='Catalog type of the Image service.'),
    cfg.StrOpt('region',
               default='',
               help="The image region name to use. If empty, the value "
                    "of identity.region is used instead. If no such region "
                    "is found in the service catalog, the first found one is "
                    "used."),
    cfg.StrOpt('http_image',
               default='http://download.cirros-cloud.net/0.3.1/'
               'cirros-0.3.1-x86_64-uec.tar.gz',
               help='http accessible image')
]

image_feature_group = cfg.OptGroup(name='image-feature-enabled',
                                   title='Enabled image service features')

ImageFeaturesGroup = [
    cfg.BoolOpt('api_v2',
                default=True,
                help="Is the v2 image API enabled"),
    cfg.BoolOpt('api_v1',
                default=True,
                help="Is the v1 image API enabled"),
]

network_group = cfg.OptGroup(name='network',
                             title='Network Service Options')

NetworkGroup = [
    cfg.StrOpt('catalog_type',
               default='network',
               help='Catalog type of the Neutron service.'),
    cfg.StrOpt('region',
               default='',
               help="The network region name to use. If empty, the value "
                    "of identity.region is used instead. If no such region "
                    "is found in the service catalog, the first found one is "
                    "used."),
    cfg.StrOpt('tenant_network_cidr',
               default="10.100.0.0/16",
               help="The cidr block to allocate tenant networks from"),
    cfg.IntOpt('tenant_network_mask_bits',
               default=28,
               help="The mask bits for tenant networks"),
    cfg.BoolOpt('tenant_networks_reachable',
                default=False,
                help="Whether tenant network connectivity should be "
                     "evaluated directly"),
    cfg.StrOpt('public_network_id',
               default="",
               help="Id of the public network that provides external "
                    "connectivity"),
    cfg.StrOpt('public_router_id',
               default="",
               help="Id of the public router that provides external "
                    "connectivity"),
]

volume_group = cfg.OptGroup(name='volume',
                            title='Block Storage Options')

VolumeGroup = [
    cfg.IntOpt('build_interval',
               default=10,
               help='Time in seconds between volume availability checks.'),
    cfg.IntOpt('build_timeout',
               default=300,
               help='Timeout in seconds to wait for a volume to become'
                    'available.'),
    cfg.StrOpt('catalog_type',
               default='volume',
               help="Catalog type of the Volume Service"),
    cfg.StrOpt('region',
               default='',
               help="The volume region name to use. If empty, the value "
                    "of identity.region is used instead. If no such region "
                    "is found in the service catalog, the first found one is "
                    "used."),
    cfg.StrOpt('backend1_name',
               default='BACKEND_1',
               help="Name of the backend1 (must be declared in cinder.conf)"),
    cfg.StrOpt('backend2_name',
               default='BACKEND_2',
               help="Name of the backend2 (must be declared in cinder.conf)"),
    cfg.StrOpt('storage_protocol',
               default='iSCSI',
               help='Backend protocol to target when creating volume types'),
    cfg.StrOpt('vendor_name',
               default='Open Source',
               help='Backend vendor to target when creating volume types'),
    cfg.StrOpt('disk_format',
               default='raw',
               help='Disk format to use when copying a volume to image'),
]

volume_feature_group = cfg.OptGroup(name='volume-feature-enabled',
                                    title='Enabled Cinder Features')

VolumeFeaturesGroup = [
    cfg.BoolOpt('multi_backend',
                default=False,
                help="Runs Cinder multi-backend test (requires 2 backends)")
]


object_storage_group = cfg.OptGroup(name='object-storage',
                                    title='Object Storage Service Options')

ObjectStoreGroup = [
    cfg.StrOpt('catalog_type',
               default='object-store',
               help="Catalog type of the Object-Storage service."),
    cfg.StrOpt('region',
               default='',
               help="The object-storage region name to use. If empty, the "
                    "value of identity.region is used instead. If no such "
                    "region is found in the service catalog, the first found "
                    "one is used."),
    cfg.IntOpt('container_sync_timeout',
               default=120,
               help="Number of seconds to time on waiting for a container"
                    "to container synchronization complete."),
    cfg.IntOpt('container_sync_interval',
               default=5,
               help="Number of seconds to wait while looping to check the"
                    "status of a container to container synchronization"),
    cfg.StrOpt('operator_role',
               default='Member',
               help="Role to add to users created for swift tests to "
                    "enable creating containers"),
]

object_storage_feature_group = cfg.OptGroup(
    name='object-storage-feature-enabled',
    title='Enabled object-storage features')

ObjectStoreFeaturesGroup = [
    cfg.BoolOpt('container_quotas',
                default=True,
                help="Set to True if the Container Quota middleware "
                     "is enabled"),
    cfg.BoolOpt('accounts_quotas',
                default=True,
                help="Set to True if the Account Quota middleware is enabled"),
    cfg.BoolOpt('crossdomain',
                default=True,
                help="Set to True if the Crossdomain middleware is enabled"),
    cfg.BoolOpt('tempurl',
                default=True,
                help="Set to True if the TempURL middleware is enabled"),
]


orchestration_group = cfg.OptGroup(name='orchestration',
                                   title='Orchestration Service Options')

OrchestrationGroup = [
    cfg.StrOpt('catalog_type',
               default='orchestration',
               help="Catalog type of the Orchestration service."),
    cfg.StrOpt('region',
               default='',
               help="The orchestration region name to use. If empty, the "
                    "value of identity.region is used instead. If no such "
                    "region is found in the service catalog, the first found "
                    "one is used."),
    cfg.BoolOpt('allow_tenant_isolation',
                default=False,
                help="Allows test cases to create/destroy tenants and "
                     "users. This option enables isolated test cases and "
                     "better parallel execution, but also requires that "
                     "OpenStack Identity API admin credentials are known."),
    cfg.IntOpt('build_interval',
               default=1,
               help="Time in seconds between build status checks."),
    cfg.IntOpt('build_timeout',
               default=300,
               help="Timeout in seconds to wait for a stack to build."),
    cfg.StrOpt('instance_type',
               default='m1.micro',
               help="Instance type for tests. Needs to be big enough for a "
                    "full OS plus the test workload"),
    cfg.StrOpt('image_ref',
               default=None,
               help="Name of heat-cfntools enabled image to use when "
                    "launching test instances."),
    cfg.StrOpt('keypair_name',
               default=None,
               help="Name of existing keypair to launch servers with."),
    cfg.IntOpt('max_template_size',
               default=524288,
               help="Value must match heat configuration of the same name."),
]


dashboard_group = cfg.OptGroup(name="dashboard",
                               title="Dashboard options")

DashboardGroup = [
    cfg.StrOpt('dashboard_url',
               default='http://localhost/',
               help="Where the dashboard can be found"),
    cfg.StrOpt('login_url',
               default='http://localhost/auth/login/',
               help="Login page for the dashboard"),
]


boto_group = cfg.OptGroup(name='boto',
                          title='EC2/S3 options')
BotoGroup = [
    cfg.StrOpt('ec2_url',
               default="http://localhost:8773/services/Cloud",
               help="EC2 URL"),
    cfg.StrOpt('s3_url',
               default="http://localhost:8080",
               help="S3 URL"),
    cfg.StrOpt('magnetodb_url',
               default="http://localhost:8080",
               help="MagnetoDB URL"),
    cfg.StrOpt('region',
               default='',
               help="The key-value-storage region name to use. If empty, the "
                    "value of identity.region is used instead. If no such "
                    "region is found in the service catalog, the first found "
                    "one is used."),
    cfg.StrOpt('aws_secret',
               default=None,
               help="AWS Secret Key",
               secret=True),
    cfg.StrOpt('aws_access',
               default=None,
               help="AWS Access Key"),
    cfg.StrOpt('s3_materials_path',
               default="/opt/stack/devstack/files/images/"
                       "s3-materials/cirros-0.3.0",
               help="S3 Materials Path"),
    cfg.StrOpt('ari_manifest',
               default="cirros-0.3.0-x86_64-initrd.manifest.xml",
               help="ARI Ramdisk Image manifest"),
    cfg.StrOpt('ami_manifest',
               default="cirros-0.3.0-x86_64-blank.img.manifest.xml",
               help="AMI Machine Image manifest"),
    cfg.StrOpt('aki_manifest',
               default="cirros-0.3.0-x86_64-vmlinuz.manifest.xml",
               help="AKI Kernel Image manifest"),
    cfg.StrOpt('instance_type',
               default="m1.tiny",
               help="Instance type"),
    cfg.IntOpt('http_socket_timeout',
               default=10,
               help="boto Http socket timeout"),
    cfg.IntOpt('num_retries',
               default=1,
               help="boto num_retries on error"),
    cfg.IntOpt('build_timeout',
               default=60,
               help="Status Change Timeout"),
    cfg.IntOpt('build_interval',
               default=1,
               help="Status Change Test Interval"),
]

stress_group = cfg.OptGroup(name='stress', title='Stress Test Options')

StressGroup = [
    cfg.StrOpt('nova_logdir',
               default=None,
               help='Directory containing log files on the compute nodes'),
    cfg.IntOpt('max_instances',
               default=16,
               help='Maximum number of instances to create during test.'),
    cfg.StrOpt('controller',
               default=None,
               help='Controller host.'),
    # new stress options
    cfg.StrOpt('target_controller',
               default=None,
               help='Controller host.'),
    cfg.StrOpt('target_ssh_user',
               default=None,
               help='ssh user.'),
    cfg.StrOpt('target_private_key_path',
               default=None,
               help='Path to private key.'),
    cfg.StrOpt('target_logfiles',
               default=None,
               help='regexp for list of log files.'),
    cfg.IntOpt('log_check_interval',
               default=60,
               help='time (in seconds) between log file error checks.'),
    cfg.IntOpt('default_thread_number_per_action',
               default=4,
               help='The number of threads created while stress test.')
]


cli_group = cfg.OptGroup(name='cli', title="cli Configuration Options")

CLIGroup = [
    cfg.BoolOpt('enabled',
                default=True,
                help="enable cli tests"),
    cfg.StrOpt('cli_dir',
               default='/usr/local/bin',
               help="directory where python client binaries are located"),
    cfg.BoolOpt('has_manage',
                default=True,
                help=("Whether the tempest run location has access to the "
                      "*-manage commands. In a pure blackbox environment "
                      "it will not.")),
    cfg.IntOpt('timeout',
               default=15,
               help="Number of seconds to wait on a CLI timeout"),
]


scenario_group = cfg.OptGroup(name='scenario', title='Scenario Test Options')

ScenarioGroup = [
    cfg.StrOpt('img_dir',
               default='/opt/stack/new/devstack/files/images/'
               'cirros-0.3.1-x86_64-uec',
               help='Directory containing image files'),
    cfg.StrOpt('ami_img_file',
               default='cirros-0.3.1-x86_64-blank.img',
               help='AMI image file name'),
    cfg.StrOpt('ari_img_file',
               default='cirros-0.3.1-x86_64-initrd',
               help='ARI image file name'),
    cfg.StrOpt('aki_img_file',
               default='cirros-0.3.1-x86_64-vmlinuz',
               help='AKI image file name'),
    cfg.StrOpt('ssh_user',
               default='cirros',
               help='ssh username for the image file'),
    cfg.IntOpt(
        'large_ops_number',
        default=0,
        help="specifies how many resources to request at once. Used "
        "for large operations testing.")
]


service_available_group = cfg.OptGroup(name="service_available",
                                       title="Available OpenStack Services")

ServiceAvailableGroup = [
    cfg.BoolOpt('cinder',
                default=True,
                help="Whether or not cinder is expected to be available"),
    cfg.BoolOpt('neutron',
                default=False,
                help="Whether or not neutron is expected to be available"),
    cfg.BoolOpt('glance',
                default=True,
                help="Whether or not glance is expected to be available"),
    cfg.BoolOpt('swift',
                default=True,
                help="Whether or not swift is expected to be available"),
    cfg.BoolOpt('nova',
                default=True,
                help="Whether or not nova is expected to be available"),
    cfg.BoolOpt('heat',
                default=False,
                help="Whether or not Heat is expected to be available"),
    cfg.BoolOpt('ceilometer',
                default=True,
                help="Whether or not Ceilometer is expected to be available"),
    cfg.BoolOpt('horizon',
                default=True,
                help="Whether or not Horizon is expected to be available"),
    # TODO Added by yyekovenko
    cfg.BoolOpt('magnetodb',
                default=True,
                help="Whether or not MagnetoDB is expected to be available"),
]

debug_group = cfg.OptGroup(name="debug",
                           title="Debug System")

DebugGroup = [
    cfg.BoolOpt('enable',
                default=True,
                help="Enable diagnostic commands"),
]


magnetodb_group = cfg.OptGroup(name="magnetodb",
                               title="Key-Value storage options")

MagnetoDBGroup = [
    cfg.StrOpt('service_type',
               default="kv-storage",
               help="The name of the MagnetoDB service type"),
]


@singleton
class TempestConfig:
    """Provides OpenStack configuration information."""

    # TODO(yyekovenko) Temporary change of path to file
    # DEFAULT_CONFIG_DIR = os.path.join(
    #    os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
    #    "etc")

    DEFAULT_CONFIG_DIR = os.path.dirname(__file__)

    DEFAULT_CONFIG_FILE = "tempest.conf"

    def __init__(self):
        """Initialize a configuration from a conf directory and conf file."""
        config_files = []
        failsafe_path = "/etc/tempest/" + self.DEFAULT_CONFIG_FILE

        # Environment variables override defaults...
        conf_dir = os.environ.get('TEMPEST_CONFIG_DIR',
                                  self.DEFAULT_CONFIG_DIR)
        conf_file = os.environ.get('TEMPEST_CONFIG', self.DEFAULT_CONFIG_FILE)

        path = os.path.join(conf_dir, conf_file)

        if not (os.path.isfile(path) or
                'TEMPEST_CONFIG_DIR' in os.environ or
                'TEMPEST_CONFIG' in os.environ):
            path = failsafe_path

        if not os.path.exists(path):
            msg = "Config file %s not found" % path
            print(RuntimeError(msg), file=sys.stderr)
        else:
            config_files.append(path)

        cfg.CONF([], project='tempest', default_config_files=config_files)
        logging.setup('tempest')
        LOG = logging.getLogger('tempest')
        LOG.info("Using tempest config file %s" % path)

        register_opt_group(cfg.CONF, compute_group, ComputeGroup)
        register_opt_group(cfg.CONF, compute_features_group,
                           ComputeFeaturesGroup)
        register_opt_group(cfg.CONF, identity_group, IdentityGroup)
        register_opt_group(cfg.CONF, image_group, ImageGroup)
        register_opt_group(cfg.CONF, image_feature_group, ImageFeaturesGroup)
        register_opt_group(cfg.CONF, network_group, NetworkGroup)
        register_opt_group(cfg.CONF, volume_group, VolumeGroup)
        register_opt_group(cfg.CONF, volume_feature_group,
                           VolumeFeaturesGroup)
        register_opt_group(cfg.CONF, object_storage_group, ObjectStoreGroup)
        register_opt_group(cfg.CONF, object_storage_feature_group,
                           ObjectStoreFeaturesGroup)
        register_opt_group(cfg.CONF, orchestration_group, OrchestrationGroup)
        register_opt_group(cfg.CONF, dashboard_group, DashboardGroup)
        register_opt_group(cfg.CONF, boto_group, BotoGroup)
        register_opt_group(cfg.CONF, compute_admin_group, ComputeAdminGroup)
        register_opt_group(cfg.CONF, stress_group, StressGroup)
        register_opt_group(cfg.CONF, scenario_group, ScenarioGroup)
        register_opt_group(cfg.CONF, cli_group, CLIGroup)
        register_opt_group(cfg.CONF, service_available_group,
                           ServiceAvailableGroup)
        register_opt_group(cfg.CONF, debug_group, DebugGroup)
        register_opt_group(cfg.CONF, magnetodb_group, MagnetoDBGroup)
        self.compute = cfg.CONF.compute
        self.compute_feature_enabled = cfg.CONF['compute-feature-enabled']
        self.identity = cfg.CONF.identity
        self.images = cfg.CONF.image
        self.image_feature_enabled = cfg.CONF['image-feature-enabled']
        self.network = cfg.CONF.network
        self.volume = cfg.CONF.volume
        self.volume_feature_enabled = cfg.CONF['volume-feature-enabled']
        self.object_storage = cfg.CONF['object-storage']
        self.object_storage_feature_enabled = cfg.CONF[
            'object-storage-feature-enabled']
        self.orchestration = cfg.CONF.orchestration
        self.dashboard = cfg.CONF.dashboard
        self.boto = cfg.CONF.boto
        self.cli = cfg.CONF.cli
        self.compute_admin = cfg.CONF['compute-admin']
        self.stress = cfg.CONF.stress
        self.scenario = cfg.CONF.scenario
        self.service_available = cfg.CONF.service_available
        self.debug = cfg.CONF.debug
        self.magnetodb = cfg.CONF.magnetodb
        if not self.compute_admin.username:
            self.compute_admin.username = self.identity.admin_username
            self.compute_admin.password = self.identity.admin_password
            self.compute_admin.tenant_name = self.identity.admin_tenant_name
